import numpy as np
import scipy.signal
from scipy import interpolate

from .base import Layer, Phi, Parameter
from .tools import require_shape, pop_shape
from nems.registry import layer
from nems.distributions import Normal, HalfNormal
from nems.tools.arrays import broadcast_axes


class FiniteImpulseResponse(Layer):

    def __init__(self, stride=1, **kwargs):
        """Convolve linear filter(s) with input.

        Parameters
        ----------
        shape : N-tuple
            Determines the shape of `FIR.coefficients`. Axes should be:
            (T time bins, C input channels (rank),  ..., N output channels)
            where only the first two dimensions are required. Aside from the
            time and filter axes (index 0 and -1, respectively), the size of
            each dimension must match the size of the input's dimensions.

            If only two dimensions are present, a singleton dimension will be
            appended to represent a single output. For higher-dimensional data,
            users are responsible for adding this singleton dimension if needed.

        See also
        --------
        nems.layers.base.Layer

        Examples
        --------
        >>> fir = FiniteImpulseResponse(shape=(15,4))   # (time, input channels)
        >>> weighted_input = np.random.rand(10000, 4)   # (time, channels)
        >>> out = fir.evaluate(weighted_input)
        >>> out.shape
        (10000, 1)

        # strf alias
        >>> fir = STRF(shape=(25, 18))                   # full-rank STRF
        >>> spectrogram = np.random.rand(10000,18)
        >>> out = fir.evaluate(spectrogram)
        >>> out.shape
        (10000, 1)

        # FIR alias                                     
        >>> fir = FIR(shape=(25, 4, 100))               # rank 4, 100 filters
        >>> spectrogram = np.random.rand(10000,4)
        >>> out = fir.evaluate(spectrogram)
        >>> out.shape
        (10000, 1, 100)

        """
        require_shape(self, kwargs, minimum_ndim=2)
        self.stride = stride
        super().__init__(**kwargs)


    def initial_parameters(self):
        """Get initial values for `FIR.parameters`.
        
        Layer parameters
        ----------------
        coefficients : ndarray
            Shape matches `FIR.shape`.
            Prior:  Normal(mean=0, sd=1/size)
            Bounds: (-np.inf, np.inf)

        Returns
        -------
        nems.layers.base.Phi

        """
        mean = np.full(shape=self.shape, fill_value=0.0)
        sd = np.full(shape=self.shape, fill_value=1/np.prod(self.shape))
        # TODO: May be more appropriate to make this a hard requirement, but
        #       for now this should stop tiny filter sizes from causing errors.
        if mean.shape[0] > 2:
            mean[1, :] = 2/np.prod(self.shape)
            mean[2, :] = -1/np.prod(self.shape)
        prior = Normal(mean, sd)

        coefficients = Parameter(name='coefficients', shape=self.shape,
                                 prior=prior)
        return Phi(coefficients)

    @property
    def coefficients(self):
        """Filter that will be convolved with input.
        
        Re-parameterized subclasses should overwrite this so that `evaluate`
        doesn't need to change.

        Returns
        -------
        coefficients : ndarray
            coefficients.shape = WeightChannels.shape
        
        """
        return self.parameters['coefficients'].values

    def evaluate(self, input):
        """Convolve `FIR.coefficients` with input."""

        # Flip rank, any other dimensions except time & number of outputs.
        coefficients = self._reshape_coefficients()
        # Match number of outputs in input and coefficients by broadcasting.
        input, coefficients = self._broadcast(input, coefficients)
        # Prepend zeros.
        padding = self._get_filter_padding(input, coefficients)
        input_with_padding = np.pad(input, padding)

        # Convolve each filter with the corresponding input channel.
        outputs = []
        n_filters = coefficients.shape[-1]
        for i in range(n_filters):
            y = scipy.signal.convolve(
                input_with_padding[...,i], coefficients[...,i], mode='valid'
                )
            outputs.append(y)

        # Concatenate on n_outputs axis
        output = np.stack(outputs, axis=-1)
        # Squeeze out rank dimension
        output = np.squeeze(output, axis=1)
        if self.stride > 1:
            output = output[::self.stride, ...]
        return output

    def _reshape_coefficients(self):
        """Get `coefficients` in the format needed for `evaluate`."""
        coefficients = self.coefficients
        if coefficients.ndim == 2:
            # Add a dummy filter/output axis
            coefficients = coefficients[..., np.newaxis]

        # Coefficients are applied "backwards" (convolution) relative to how
        # they are specified (filter), so have to flip all dimensions except
        # time and number of filters/outputs.
        flipped_axes = [1]  # Always flip rank
        other_dims = coefficients.shape[2:-1]
        for i, d in enumerate(other_dims):
            # Also flip any additional dimensions
            flipped_axes.append(i+2)
        coefficients = np.flip(coefficients, axis=flipped_axes)

        return coefficients

    def _broadcast(self, input, coefficients):
        """Internal for `evaluate`."""
        # Add axis for n output channels to input if one doesn't exist.
        # NOTE: This will only catch a missing output dimension for 2D data.
        #       For higher-dimensional data, the output dimension needs to be
        #       specified by users.
        if input.ndim < 3:
            input = input[..., np.newaxis]

        if input.shape[-1] < coefficients.shape[-1]:
            try:
                input = broadcast_axes(input, coefficients, axis=-1)
            except ValueError:
                raise TypeError(
                    "Last dimension of FIR input must match last dimension of "
                    "coefficients, or one must be broadcastable to the other."
                    )
        elif coefficients.shape[-1] < input.shape[-1]:
            try:
                coefficients = broadcast_axes(coefficients, input, axis=-1)
            except ValueError:
                raise TypeError(
                    "Last dimension of FIR input must match last dimension of "
                    "coefficients, or one must be broadcastable to the other."
                    )
        
        return input, coefficients

    def _get_filter_padding(self, input, coefficients):
        """Get zeros of correct shape to prepend to input on time axis."""

        filter_length = coefficients.shape[0]
        # Prepend 0s on time axis, no padding on other axes
        padding = [[filter_length-1, 0]] + [[0, 0]]*(input.ndim-1)

        return padding

    @layer('fir')
    def from_keyword(keyword):
        """Construct FIR (or subclass) from keyword.

        Keyword options
        ---------------
        {digit}x{digit}x ... x{digit} : N-dimensional shape.
            (time, input channels a.k.a. rank, ..., output channels) 

        See also
        --------
        Layer.from_keyword
        
        """
        kwargs = {}
        fir_class = FiniteImpulseResponse

        options = keyword.split('.')
        kwargs['shape'] = pop_shape(options)
        for op in options:
            if op.startswith('p') and op[1].isdigit():
                # Pole-zero parameterization
                fir_class = PoleZeroFIR
                fs_idx = op.index('fs')
                zeros_idx = op.index('z')
                kwargs['n_poles'] = int(op[1:zeros_idx])
                kwargs['n_zeros'] = int(op[zeros_idx+1:fs_idx])
                kwargs['fs'] = int(op[fs_idx+2:])
            elif op.startswith('s'):
                kwargs['stride'] = int(op[1:])
        fir = fir_class(**kwargs)

        return fir
    
    def as_tensorflow_layer(self, input_shape, **kwargs):
        """Convert FiniteImpulseResponse to a TensorFlow Keras Layer.
        
        Parameters
        ----------
        inputs : tf.Tensor or np.ndarray
            Initial input to Layer, supplied by TensorFlowBackend during model
            building.
        
        Returns
        -------
        FiniteImpulseResponseTF
        
        """

        import tensorflow as tf
        from nems.backends.tf import NemsKerasLayer

        old_c = self.parameters['coefficients']
        coefficients = self.coefficients
        if coefficients.ndim == 2:
            # Add a dummy filter/output axis
            coefficients = coefficients[..., np.newaxis]
        new_c = np.flip(coefficients, axis=0)
        filter_width, rank, _ = new_c.shape
        if new_c.ndim > 3:
            raise NotImplementedError(
                "FIR TF implementation currently only works for 2D data."
                )
        new_values = {'coefficients': new_c}  # override Parameter.values

        # Define broadcasting behavior for inputs and coefficients based on
        # input_shape and new_c.shape.
        broadcast_inputs, broadcast_coefficients, n_outputs = \
            self._define_tf_broadcasting(
                tf, input_shape, new_c
                )
        # Define convolution operation, depends on whether a GPU is available.
        convolve = self._define_tf_convolution(
            tf, filter_width, rank, n_outputs
            )

        class FiniteImpulseResponseTF(NemsKerasLayer):
            def weights_to_values(self):
                c = self.parameter_values['coefficients']
                unflipped = np.flip(c, axis=0)  # Undo flip time
                unshaped = np.reshape(unflipped, old_c.shape)

                return {'coefficients': unshaped}

            def call(self, inputs):
                # This will add an extra dim if there is no output dimension.
                input_width = tf.shape(inputs)[1]
                # Broadcast output shape if needed.
                inputs = broadcast_inputs(inputs)
                coefficients = broadcast_coefficients(self.coefficients)
                # Make None shape explicit
                rank_4 = tf.reshape(inputs, [-1, input_width, rank, n_outputs])
                return convolve(rank_4, coefficients)

        return FiniteImpulseResponseTF(self, new_values=new_values, **kwargs)


    def _define_tf_broadcasting(self, tf, input_shape, new_c):
        """Internal for `as_tensorflow_layer`.
        
        Builds `broadcast_inputs` and `broadcast_coefficients` for use inside
        `call` method.

        Parameters
        ----------
        tf : package
            Reference to imported TensorFlow package.
        input_shape : tuple
            Shape of inputs.
        new_c : np.ndarray
            Reshaped coefficients.

        Returns
        -------
        broadcast_inputs : function
        broadcast_coefficients : function
        n_outputs : int
            Number of broadcasted outputs.

        """

        # Fake input to set up correct broadcasting behavior.
        # Only the number of outputs matters, this drops the batch dimension.
        # TODO: This might mess up with multiple batches similar to WC?
        #       Need to check if list?
        fake_inputs = np.empty(shape=input_shape[1:])
        new_inputs, broadcast_c = self._broadcast(fake_inputs, new_c)
        new_coefs_shape = list(new_c.shape[:-1]) + [broadcast_c.shape[-1]]
        new_inputs_shape = list(new_inputs.shape)
        n_outputs = new_coefs_shape[-1]

        if new_inputs_shape[-1] > input_shape[-1]:
            # If output dimension increased, then TF needs to broadcast output
            # dimension of input in call function.
            if new_inputs.ndim > fake_inputs.ndim:
                # A singleton output dimension needs to be appended as well.
                @tf.function
                def expand_inputs(inputs):
                    return tf.expand_dims(inputs, axis=-1)
            else:
                @tf.function
                def expand_inputs(inputs):
                    return inputs

            @tf.function()
            def broadcast_inputs(inputs):
                # Convert None batch shape to int, add singleton output dim
                # if needed. Then broadcast outputs.
                batch_size = tf.keras.backend.shape(inputs)[0]
                shape = [batch_size] + new_inputs_shape
                return tf.broadcast_to(expand_inputs(inputs), shape)
        else:
            # Otherwise, don't need to do anything to inputs.
            @tf.function
            def broadcast_inputs(inputs):
                # This will still add a singleton output dim if needed.
                #return tf.reshape(inputs, new_inputs_shape)
                return inputs

        if new_coefs_shape[-1] > new_c.shape[-1]:
            # Coefficients outputs increased, need to broadcast coefs in call.
            @tf.function
            def broadcast_coefficients(coefficients):
                return tf.broadcast_to(coefficients, new_coefs_shape)
        else:
            # Otherwise, don't need to do anything to coefficients.
            @tf.function
            def broadcast_coefficients(coefficients): return coefficients
        
        return broadcast_inputs, broadcast_coefficients, n_outputs

    def _define_tf_convolution(self, tf, filter_width, rank, n_outputs):
        """Internal for `as_tensorflow_layer`.
        
        Builds `convolution` function for use in `call` method.

        Parameters
        ----------
        tf : package
            Reference to imported TensorFlow package.
        filter_width, rank, n_outputs : coefficient shape components 

        Returns
        -------
        convolution : function
        
        """

        num_gpus = len(tf.config.list_physical_devices('GPU'))
        stride = self.stride
        if num_gpus == 0:
            # Use CPU-compatible (but slower) version.
            @tf.function
            def convolve(inputs, coefficients):
                # Reorder coefficients to shape (n outputs, time, rank, 1)
                new_coefs = tf.expand_dims(
                    tf.transpose(coefficients, [2, 0, 1]), -1
                    )
                padded_input = tf.pad(
                    inputs, [[0, 0], [filter_width-1, 0], [0, 0], [0, 0]]
                    )
                # Reorder input to shape (n outputs, batch, time, rank)
                x = tf.transpose(padded_input, [3, 0, 1, 2])
                fn = lambda t: tf.cast(tf.nn.conv1d(  # TODO: don't like forcing dtype here
                    t[0], t[1], stride=stride, padding='VALID'
                    ), tf.float64)
                # Apply convolution for each output
                y = tf.map_fn(
                    fn=fn,
                    elems=(x, new_coefs),
                    fn_output_signature=tf.float64
                    )
                # Reorder output back to (batch, time, n outputs)
                z = tf.transpose(tf.squeeze(y, axis=3), [1, 2, 0])
                return z
        else:
            # Use GPU-only version (grouped convolutions), much faster.
            @tf.function
            def convolve(inputs, coefficients):
                input_width = tf.shape(inputs)[1]
                # Reshape will group by output before rank w/o transpose.
                transposed = tf.transpose(inputs, [0, 1, 3, 2])
                # Collapse rank and n_outputs to one dimension.
                # -1 for batch size b/c it can be None.
                reshaped = tf.reshape(
                    transposed, [-1, input_width, rank*n_outputs]
                    )
                # Prepend 0's on time axis as initial conditions for filter.
                padded_input = tf.pad(
                    reshaped, [[0, 0], [filter_width-1, 0], [0, 0]]
                    )
                # Convolve filters with input slices in groups of size `rank`.
                y = tf.nn.conv1d(
                    padded_input, coefficients, stride=stride, padding='VALID'
                    )
                return y 

        return convolve


# Aliases, STRF specifically for full-rank (but not enforced)
class FIR(FiniteImpulseResponse):
    pass
class STRF(FiniteImpulseResponse):
    pass


class PoleZeroFIR(FiniteImpulseResponse):

    def __init__(self, n_poles, n_zeros, fs, **kwargs):
        """TODO: docs.
        
        TODO: Possible to remove need for sampling rate?
        
        """
        self.n_poles = n_poles
        self.n_zeros = n_zeros
        self.fs = fs
        require_shape(self, kwargs, minimum_ndim=2, maximum_ndim=3)
        super().__init__(**kwargs)

    def initial_parameters(self):
        """TODO: docs."""

        # TODO: explain choice of priors
        rank = self.shape[1]
        if len(self.shape) == 3:
            n_filters = self.shape[2]
        else:
            n_filters = 1
        pole_set = np.array([[[0.8, -0.4, 0.1, 0.0, 0]]])[..., :self.n_poles]
        zero_set = np.array([[[0.1,  0.1, 0.1, 0.1, 0]]])[..., :self.n_zeros]

        poles_prior = Normal(
            mean = pole_set.repeat(rank, 0).repeat(n_filters, 1),
            sd = np.ones((rank, n_filters, self.n_poles))*0.3,
            )
        zeros_prior = Normal(
            mean = zero_set.repeat(rank, 0).repeat(n_filters, 1),
            sd = np.ones((rank, n_filters, self.n_zeros))*0.2,
            )
        delays_prior = HalfNormal(sd = np.ones((rank, n_filters))*0.02)
        gains_prior = Normal(
            mean = np.zeros((rank, n_filters))+0.1,
            sd = np.ones((rank, n_filters))*0.2
            )

        poles = Parameter('poles', shape=(rank, n_filters, self.n_poles),
                          prior=poles_prior, bounds=(-1, 1))
        zeros = Parameter('zeros', shape=(rank, n_filters, self.n_zeros),
                          prior=zeros_prior, bounds=(-1, 1))
        # TODO: what do the delays do exactly?
        delays = Parameter('delays', shape=(rank, n_filters),
                           prior=delays_prior, bounds=(0, np.inf))
        gains = Parameter('gains', shape=(rank, n_filters),
                          prior=gains_prior)

        return Phi(poles, zeros, delays, gains)

    @property
    def coefficients(self):
        """TODO: docs."""
        poles, zeros, delays, gains = self.get_parameter_values()

        n_taps, rank = self.shape[:2]
        if len(self.shape) == 2:
            n_filters = 1
        else:
            n_filters = self.shape[-1]

        coefficients = np.zeros((n_taps, rank, n_filters))

        # TODO: can we do this without fs?
        # TODO: explain why 5*original
        fs2 = 5*self.fs                      

        for i in range(rank):
            for j in range(n_filters):
                # TODO: rename variables, improve documentation.
                #       still don't really know what this is doing.
                t = np.arange(0, n_taps*5 + 1) / fs2
                h = scipy.signal.ZerosPolesGain(
                    zeros[i,j], poles[i,j], gains[i,j], dt=1/fs2
                    )
                tout, ir = scipy.signal.dimpulse(h, t=t)
                f = interpolate.interp1d(tout, ir[0][:,0], bounds_error=False,
                                         fill_value=0)

                tnew = np.arange(0, n_taps)/self.fs - delays[i,j] + 1/self.fs
                coefficients[:, i, j] = f(tnew)

        return coefficients

    # TODO: as_tensorflow_layer



"""A collection of miscellaneous utility functions.

TODO: Anything that used to be in nems.utils.py should go in this directory,
      with functions of similar purpose grouped in modules. Even if there's just
      one function that belongs in a solo group, it should *not* just be dumped
      in 'misc' or 'etc' or a similarly named file with other one-offs.

      Alternatively, if all utilites can be reasonably assigned to specific
      libraries, we can just get rid of this directory altogether.


Contents
--------
    `json.py` : Ensure proper json-ification of Models, Layers, etc.
                NOTE: tools from `json.py` should *not* be imported here
                (for the `nems.tools` API), as this can easily cause circular
                imports.
    `arrays.py` : Tools for manipulating NumPy arrays. Useful for working with
                  the dictionary of arrays format that `nems.Model` uses for
                  complex model fits.
    `lookup.py` : Look up functions based on strings. Mainly used to improve
                  user-friendliness when specifying cost functions or similar
                  options.

"""



"""Computational frameworks for fitting Models.

To include: SciPy (default), TensorFlow, ... (others)

NOTE: There are no public API imports here so that backend dependencies are
      only imported if used. For example, users don't need to install TensorFlow
      unless they want to fit models using `Model.fit(..., backend='tf')`.

"""

# TODO: better way to do this. For now, new backends just have to add on to
#       this if/else chain.
def get_backend(name):
    """Import and return a Backend subclass corresponding to `name`."""
    name = name.lower()
    if name == 'scipy':
        from .scipy import SciPyBackend
        backend = SciPyBackend
    elif (name == 'tf') or (name == 'tensorflow'):
        from .tf import TensorFlowBackend
        backend = TensorFlowBackend
    else:
        raise NotImplementedError(f"Unrecognized backend: {name}.")

    return backend



'''
Collects commonly-used Modules for easier import.

Ex: `from nems.modules import WeightChannels, FIR, DoubleExponential`

'''

from .nonlinearity import LevelShift, DoubleExponential, RectifiedLinear, ReLU
from .filter import FiniteImpulseResponse, FIR, STRF
from .weight_channels import WeightChannels, WeightChannelsMulti, WeightChannelsGaussian, WeightChannelsMultiGaussian
from .state import StateGain, StateDexp
from .numpy import NumPy
from .stp import ShortTermPlasticity, STP
from .algebra import SwapDims

from .base import Layer, Phi, Parameter



from scipy import stats
import numpy as np

from .base import Distribution


class Uniform(Distribution):
    """Uniform prior.

    Parameters
    ----------
    lower : scalar or ndarray
        Lower bound of distribution
    upper : scalar or ndarray
        Upper bound of distribution
    
    """

    def __init__(self, lower, upper):
        self._lower = np.asarray(lower)
        self._upper = np.asarray(upper)
        self.distribution = stats.uniform(
            loc=self._lower, scale=self._upper-self._lower
            )

    def __repr__(self):
        lower = self.value_to_string(self._lower)
        upper = self.value_to_string(self._upper)
        return 'Uniform({}, {})'.format(lower, upper)



# TODO
# From supplement doc:
# "align and combine datasets for fitting a single, large model"

# So, for example, combining PEG and A1 data to fit with a single pop model.

from nems.tools.arrays import concatenate_dicts


def combine_neurons(*datasets, neurons_keys='response', axes=-1, newaxis=False):
    """Merge multiple datasets by concatenating on neural variables.

    This is essentially a wrapper for `concatenate_dicts` (and, ultimately,
    `np.concatenate`) with some extra options parsing.

    Parameters
    ----------
    datasets : N-tuple of dict.
        Each dictionary should be in the format expected for the `input` arg of
        `nems.Model.evaluate`. That is:
            `{'key1': np.ndarray, 'key2': np.ndarray, ...}`,
        where each array has shape (T, ..., N) or (S, T, ..., N)

    """



"""Collects commonly-used core classes & functions for easier import.

Ex: `from nems import Model`

"""

from nems.models.base import Model
from nems.tools.demo_data.file_management import download_demo, load_demo



import textwrap

import numpy as np


class Backend:

    def __init__(self, nems_model, data, eval_kwargs=None, **backend_options):
        """Interface between NEMS Models and optimization libraries.

        Each Backend must implement `_build`, `_fit`, and `predict` methods.

        Parameters
        ----------
        nems_model : nems.models.base.Model
            `Backend._build` should return an "equivalent" representation of
            this Model.
        data : DataSet
            Model inputs, outputs and targets for use by `Backend._build`.
        eval_kwargs : dict
            Keyword arguments for `nems_model.evaluate`.
        backend_options : dict
            Additional options for Backend subclasses.
        
        """
        self.nems_model = nems_model
        if eval_kwargs is None: eval_kwargs = {}
        self.model = self._build(
            data, eval_kwargs=eval_kwargs, **backend_options
            )

    def _build(self, data, eval_kwargs=None, **backend_options):
        """Construct whatever object this Backend uses for fitting/predicting.
        
        Parameters
        ----------
        data : DataSet
            Model inputs, outputs and targets.
        eval_kwargs : dict
            Keyword arguments for `Backend.nems_model.evaluate`.
        backend_options : dict
            Additional options for Backend subclasses.

        Returns
        -------
        object
            Some object defined by the Backend that handles evaluating data
            and optimizing parameters (comparable to a NEMS Model).

        """
        raise NotImplementedError
        
    def _fit(self, data, eval_kwargs=None, **kwargs):
        """Call Backend.model.fit() or equivalent.
        
        This method should update parameters in-place for `Backend.model` and
        `Backend.nems_model`, and store information about optimization results
        in the returned FitResults object.

        Parameters
        ----------
        data : DataSet
            Model inputs, outputs and targets.
        eval_kwargs : dict
            Keyword arguments for `Backend.nems_model.evaluate`.
        kwargs : dict
            Subclasses can specify additional keyword arguments that will be
            supplied by `Model.fit` as `Backend._fit(..., **fitter_options)`.

        Returns
        -------
        FitResults
        
        """
        raise NotImplementedError

    def predict(self, input, **eval_kwargs):
        """Call _model.predict() or equivalent.
        
        Parameters
        ----------
        input : ndarray, dict, or DataSet
            See `nems.models.base.Model.evaluate` for details.
        eval_kwargs : dict
            Additional keyword arguments for `Backend.nems_model.evaluate`.

        Returns
        -------
        prediction : (varies)
            Some form of data representing the model's output given `input`.
            No specific type is enforced at this time, but it should be readily
            convertible to `np.ndarray` (or a list or dict of ndarrays).
        
        """
        raise NotImplementedError


class FitResults:

    def __init__(self, initial_parameters, final_parameters, initial_error,
                 final_error, backend_name, **misc):
        """Collection of information about one fit of a NEMS Model.

        Parameters
        ----------
        initial_parameters : list or np.ndarray
            Parameters of `Backend.nems_model` prior to fitting, formatted as
            a flattened list or 1D ndarray.
        final_parameters : list of np.ndarray
            Parameters of `Backend.nems_model` after fitting.
        initial_error : float
            `cost_function(input, target)` prior to fitting.
        final_error : float
            `cost_function(input, target)` after fitting.
        backend_name : str
            Name indicating which Backend was used to fit the model.
        misc : dict
            Additional fit-related information to store.

        Attributes
        ----------
        initial_parameters : np.ndarray
        final_parameters : np.ndarray
        initial_error : float
        final_error : float
        backend : str
        n_parameters : int
            Number of parameters in the model.
        n_parameters_changed : int
            Number of indices `i` for which `initial_parameters[i]` does not
            match `final_parameters[i]`.
        
        """
        self.initial_parameters = np.array(initial_parameters)
        self.final_parameters = np.array(final_parameters)
        self.initial_error = initial_error
        self.final_error = final_error
        self.n_parameters = self.initial_parameters.size
        self.n_parameters_changed = np.sum(
            (self.initial_parameters != self.final_parameters)
            )
        self.backend = backend_name
        self.misc = misc

    def __repr__(self):
        attrs = self.__dict__.copy()
        x0 = attrs.pop('initial_parameters')  # TODO: maybe not worth printing?
        x1 = attrs.pop('final_parameters')
        misc = attrs.pop('misc')

        string = "Fit Results:\n"
        string += "="*11 + "\n"
        for k, v in attrs.items():
            string += f"{k}: {v}\n"
        string += f"Misc:\n"
        for k, v in misc.items():
            string += '-'*11 + '\n'
            string += f'{k}:\n{v}\n'
        string += "="*11

        return string

    def to_json(self):
        args = [
            self.initial_parameters,
            self.final_parameters,
            self.initial_error,
            self.final_error,
            self.backend,
        ]

        kwargs = self.misc.copy()
        if self.backend == 'SciPy':
            h = kwargs['scipy_fit_result'].hess_inv
            kwargs['scipy_fit_result'].hess_inv = h.todense()  # np array

        return {'args': args, 'kwargs': kwargs}

    @classmethod
    def from_json(cls, json):
        if json['args'][-1] == 'SciPy':  # backend
            # Convert dict back to object
            class ReloadedOptimizeResult:
                pass
            o = ReloadedOptimizeResult()

            for k, v in json['kwargs']['scipy_fit_result'].items():
                setattr(o, k, v)
            json['kwargs']['scipy_fit_result'] = o

        return FitResults(*json['args'], **json['kwargs'])



import numpy as np
from numba import njit

from nems.registry import layer
from nems.layers.base import Layer, Phi, Parameter
from nems.layers.tools import require_shape, pop_shape
from nems.tools.arrays import one_or_more_negative
from nems.distributions import Normal, HalfNormal


class ShortTermPlasticity(Layer):

    def __init__(self, quick_eval=False, fs=100.0, crosstalk=0, dep_only=False,
                 chunksize=5, **kwargs):
        """TODO: docs

        TODO: additional context.
        Need @SVD's help filling in all these docs, some parameter renaming
        would likely be helpful as well.
        
        Parameters
        ----------
        shape : tuple.
            TODO: what's the shape syntax? Only one dim, multiple? What do
                  the value(s) represent?
        quick_eval : bool; default=True.
            Determines which implementation of the STP algorithm to use.
            TODO: explain why some one might want to use `quick_eval=False`
            (or if there's no good reason, just get rid of it). I think for now
            it's just b/c faciliatation isn't implemented?
        fs : float; default=100.0.
            Frequency of sampling, in Hz. Used to determine reasonable bounds
            for parameters.
        crosstalk : int; default=0;
            TODO: explain what this does.
            TODO: or remove if this isn't used any more?
        dep_only : bool; default=False.
            TODO: explain what this does.
        chunksize : int; default=5.
            TODO: explain what this does.
            NOTE: used to be in units of seconds, now it should be specified
                  in units of bins.

        """

        require_shape(self, kwargs, minimum_ndim=1)
        self.quick_eval = quick_eval
        self.fs = fs
        self.crosstalk = crosstalk
        self.dep_only = dep_only
        self.chunksize = chunksize
        super().__init__(**kwargs)


    def initial_parameters(self):
        """TODO: docs
        
        Layer parameters
        ----------------
        u : np.ndarray.
            TODO: explain purpose.
            TODO: expected shape?
            Prior: HalfNormal(sd=0.1)
            Bounds: (0.0001, np.inf)   TODO: should this use epsilon instead?
        tau : np.ndarray.
            TODO: explain purpose.
            TODO: expected shape?
            Prior: HalfNormal(sd=0.1)
            Bounds: (1e-6, np.inf)     TODO: should this use epsilon instead?
        
        """


        u0, tau0 = self.seconds_to_bins(self.fs, 0.1, 0.1)     # 0.1 frac, 100 ms
        _, tau_min = self.seconds_to_bins(self.fs, 0.1, 0.001) #   1 ms

        tau_sd = np.full(shape=self.shape, fill_value=tau0)
        tau_prior = HalfNormal(tau_sd)
        u_sd = np.full(shape=self.shape, fill_value=u0)
        u_prior = HalfNormal(u_sd)

        tau_bounds = (tau_min, np.inf)
        if self.dep_only or self.quick_eval:
            u_bounds = (1e-6, np.inf)
        else:
            u_bounds = (-np.inf, np.inf)

        u = Parameter('u', shape=self.shape, prior=u_prior, bounds=u_bounds)
        tau = Parameter('tau', shape=self.shape, prior=tau_prior,
                        bounds=tau_bounds)

        return Phi(u, tau)


    def bins_to_seconds(self, fs, u=None, tau=None):
        """Get parameter values in units of seconds based on sampling rate."""
        
        if u is None: u = self.get_parameter_values('u')
        if tau is None: tau = self.get_parameter_values('tau')
        u_seconds = (u/100)*fs
        tau_seconds = tau/fs

        return u_seconds, tau_seconds

    def seconds_to_bins(self, fs, u, tau):
        """Get parameter values in units of bins based on sampling rate."""
        u_bins = (u/fs)*100
        tau_bins = tau*fs

        return u_bins, tau_bins


    # Authors: SVD, Menoua K.
    # Revision by: JRP.
    def evaluate(self, input):
        """TODO: docs

        TODO: is there a succint equation we can put here to represent what
              STP is doing?

        TODO: brief explanation of what the Layer is doing to inputs.

        """
        
        if one_or_more_negative(input):
            raise ValueError(
                "STP only supports inputs with all non-negative values. "
                "We recommend normalizing inputs between 0 and 1, and/or "
                "preceeding STP with a layer that guarantees positive outputs."
            )

        try:
            n_times, n_channels = input.shape
        except ValueError as e:
            if "not enough values to unpack" in str(e):
                raise ValueError(
                    "STP only supports 2D inputs with shape (T, N), where T "
                    "is the number of time bins and N is the number of output "
                    "channels."
                )

        u, tau = self.get_parameter_values()
        # TODO: temporary fudge factor to deal with scaling priors, since
        #       half-normal can only specify sd. remove this after implementing
        #       a proper parameter scaling system.
        tau *= 100

        # TODO: Refactor if still using, otherwise remove.
        # if self.crosstalk:
        #     # assumes dim of u is 1 !
        #     tstim = np.mean(input, axis=0, keepdims=True)

        if self.quick_eval:
            a = 1 / tau[np.newaxis, ...]
            x = u[np.newaxis, ...] * input
            nonzero = (x > 0)
            # Values that aren't set get multiplied by 0, so value of empty
            # entries doesn't matter and this is much faster than initializing
            # to ones.
            depression_per_bin = np.empty_like(input) 
            x0, imu0 = 0.0, 0.0

            i = 0
            j = self.chunksize
            while i < (n_times - 1):
                # Approximates analytical solution to differential equation:
                # TODO: fill in diffeq (or maybe put this in docstring for
                #       quick_eval kwarg instead).
                xi = x[i:j, :]
                x1 = xi[:1, :]

                ix = self._cumulative_integral_trapz(a + xi) + a + (x0 + x1)/2
                mu = np.exp(ix)
                mu1 = mu[:1, :]
                imu = self._cumulative_integral_trapz(mu * xi) \
                        + (x0 + mu1*x1)/2 + imu0

                _nonzero = nonzero[i:j, :]
                depression_per_bin[i:j, :][_nonzero] = \
                        1 - imu[_nonzero] / mu[_nonzero]
                x0 = xi[-1:, :]
                imu0 = imu[-1:, :] / mu[-1:, :]

                i += self.chunksize
                j += self.chunksize

            # Shift depression forward in time by one to allow STP to kick in
            # after the stimulus changes.
            out = np.multiply(
                input, np.concatenate(
                    # Oddly enough, zeros + 1 is twice as fast as using ones
                    # for this size of array (1-6ish usually)
                    [np.zeros((1, n_channels)) + 1, depression_per_bin[:-1, :]],
                    axis=0
                    ),
                )

        else:
            out = np.empty_like(input)
            for i in range(n_channels):
                depression_per_bin = self._inner_loop(u, tau, input, i)

                # TODO: refactor if still using, otherwise remove
                # if self.crosstalk:
                #     stim_out *= np.expand_dims(td, 0)
                out[:, i] = np.multiply(
                    input[:, i], depression_per_bin,
                    )

        return out


    @staticmethod
    @njit
    def _inner_loop(u, tau, input, i):
        """TODO: docs.
        
        Internal for `evaluate`.
        
        """

        a = 1 / tau[i]
        ustim = 1.0 / tau[i] + u[i] * input[:, i]
        s = ustim.shape[0]
        td = [1]  # initialize dep state vector

        if u[i] == 0:
            # passthru, no STP, preserve stim_out = tstim
            pass
        else:
            if u[i] > 0:
                depression = True
            else:
                depression = False

            for tt in range(1, s):
                delta = a - td[-1] * ustim[tt - 1]
                next_td = td[-1] + delta

                if depression and next_td < 0:
                    next_td = 0
                elif not depression and next_td > 5:
                    # TODO: Make this a hyperparameter?
                    #       Hard-coded 5 avoids explosions, and it's big enough
                    #       that it's essentially "infinity" in biological terms.
                    next_td = 5
                td.append(next_td)

        return np.array(td)


    @staticmethod
    def _cumulative_integral_trapz(y):
        """Cumulative integral of y(x) using the trapezoid method.
        
        Compare to `scipy.integrate.cumtrapz`. This method is slightly faster
        due to supporting fewer options, with fixed parameters:
            `x = None`
            `dx = 1`
            `axis = 0`
            `initial = 0`
        
        Parameters
        ----------
        y : np.ndarray.
            Values to integrate.

        Returns
        -------
        np.ndarray
            Integrated values, with the same shape as `y`.
        
        """
        y = (y[:-1, :] + y[1:, :]) / 2.0
        y = np.concatenate([np.zeros((1, y.shape[1])), y], axis=0)
        y = np.cumsum(y, axis=0)

        return y

    @layer('stp')
    def from_keyword(keyword):
        """TODO: docs"""
        options = keyword.split('.')
        shape = pop_shape(options)

        return ShortTermPlasticity(shape=shape)

    def as_tensorflow_layer(self, **kwargs):

        """
        ported from nems0.tf.layers
        TODO: docs"""
        import tensorflow as tf
        from nems.backends.tf import NemsKerasLayer
    

        parent_x0 = self.x0
        fs = self.fs
        crosstalk = self.crosstalk
        if crosstalk:
            # TODO: Remove this after implementing crosstalk
            raise NotImplemented(
                'STP(..., crosstalk=True) is not yet supported for the '
                'TensorFlow backend.'
                )
        parent_chunksize = self.chunksize
        reset_signal = self.reset_signal

        @tf.function
        def _cumtrapz(x, dx=1., initial=0.):
            x = (x[:, :-1] + x[:, 1:]) / 2.0
            x = tf.pad(
                x, ((0, 0), (1, 0), (0, 0)), constant_values=initial
                )

            return tf.cumsum(x, axis=1) * dx


        class ShortTermPlasticityTF(NemsKerasLayer):
            def call(self, inputs):
                # TODO docs.
                # Implementation developed by Menoua K.

                # TODO: Why are different levels of precision hard-coded in
                #       different places? Ideally should change all dtype=
                #       to inputs.dtype to work with the new consistent-dtype
                #       system.
                _zero = tf.constant(0.0, dtype='float32')
                _nan = tf.constant(0.0, dtype='float32')

                s = inputs.shape
                tstim = tf.where(tf.math.is_nan(inputs), _zero, inputs)
                tstim = tf.nn.relu(tstim)

                if parent_x0 is not None:  # x0 should be tf variable to avoid retraces
                    # TODO: is this expanding along the right dim? tstim dims: (None, time, chans)
                    tstim = tstim - tf.expand_dims(parent_x0, axis=1)

                # convert a & tau units from sec to bins
                ui = tf.math.abs(tf.reshape(self.u, (1, -1))) / fs * 100
                taui = tf.math.abs(tf.reshape(self.tau, (1, -1))) * fs

                # convert chunksize from sec to bins
                chunksize = int(parent_chunksize * fs)

                if crosstalk:
                    # assumes dim of u is 1 !
                    tstim = tf.math.reduce_mean(tstim, axis=0, keepdims=True)

                ui = tf.expand_dims(ui, axis=0)
                taui = tf.expand_dims(taui, axis=0)

                a = tf.cast(1.0 / taui, 'float64')
                x = ui * tstim

                # TODO: move these outside __call__ similar to revision of scipy
                if reset_signal is None:
                    reset_times = tf.range(0, s[1] + chunksize - 1, chunksize)
                else:
                    reset_times = tf.where(reset_signal[0, :])[:, 0]
                    reset_times = tf.pad(
                        reset_times, ((0, 1),), constant_values=s[1]
                        )

                td = []
                x0, imu0 = 0.0, 0.0
                for j in range(reset_times.shape[0] - 1):
                    xi = tf.cast(
                        x[:, reset_times[j]:reset_times[j + 1], :],
                        'float64'
                        )
                    ix = _cumtrapz(a + xi, dx=1, initial=0) \
                         + a + (x0 + xi[:, :1]) / 2.0

                    mu = tf.exp(ix)
                    imu = _cumtrapz(mu * xi, dx=1, initial=0) \
                          + (x0 + mu[:, :1] * xi[:, :1]) / 2.0 + imu0

                    valid = tf.logical_and(mu > 0.0, imu > 0.0)
                    mu = tf.where(valid, mu, 1.0)
                    imu = tf.where(valid, imu, 1.0)
                    _td = 1 - tf.exp(tf.math.log(imu) - tf.math.log(mu))
                    _td = tf.where(valid, _td, 1.0)

                    x0 = xi[:, -1:]
                    imu0 = imu[:, -1:] / mu[:, -1:]
                    td.append(tf.cast(_td, 'float32'))
                td = tf.concat(td, axis=1)

                ret = tstim * tf.pad(
                        td[:, :-1, :], ((0, 0), (1, 0), (0, 0)),
                        constant_values=1.0
                        )
                ret = tf.where(tf.math.is_nan(inputs), _nan, ret)

                return ret

        return ShortTermPlasticityTF(self, **kwargs)


class STP(ShortTermPlasticity):
    pass



import numpy as np

from nems.registry import layer
from nems.visualization import plot_layer
from nems.tools.arrays import one_or_more_nan
from .phi import Phi
from .map import DataMap


# TODO: add more examples, and tests
# TODO: add option to propagate other Parameter options from Layer.__init__,
#       like `default_bounds` and `initial_value`.
class Layer:
    """Encapsulates one data-transformation step in a  Model.

    Attributes
    ----------
    subclasses : dict
        A dictionary of classes that inherit from Layer, used by `from_json`.
    state_arg : str or None; optional.
        If string, `state_arg` will be interpreted as the name of an
        argument for `Layer.evaluate`. During Model evaluation, if
        `Layer.input is None` and a `state` array is provided to
        `Model.evaluate`, then `state` will be added to other inputs as a
        keyword argument, i.e.: `layer.evaluate(*inputs, **state)`.
    _inplace_ok : bool; default=False.
        Layer.evaluate can reference this flag to determine if it can safely
        overwrite input arrays with new values to reduce memory usage within
        the scope of `evaluate`. A parent Model will set `inplace_ok=True`
        only if the output of the previous Layer will not be saved. Setting
        this attribute directly is not recommended except when debugging.

    """

    # Any subclass of Layer will be registered here, for use by
    # `Layer.from_json`
    subclasses = {}
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.subclasses[cls.__name__] = cls

    state_arg = None
    _inplace_ok = False

    def __init__(self, shape=None, input=None, output=None, parameters=None,
                 priors=None, bounds=None, name=None):
        """Encapsulates one data-transformation step of a NEMS ModelSpec.

        Layers are intended to exist as components of a parent Model by
        invoking `Model.add_layer` or `Model.__init__(layers=...)`.

        Parameters
        ----------
        shape : N-tuple of int or None; optional.
            Most Layer subclasses use this keyword argument to specify the
            shape of their Parameters. See individual subclasses for expected
            format. For built-in Layers, an error will typically be raised if
            shape is not specified.
        input : str, list, dict, or None; optional.
            Specifies which data streams should be provided as inputs by
            parent Model during evaluation, where strings refer to keys for a
            dict of arrays provided to `Model.fit`.
            If None : output of previous Layer.
            If str  : a single input array.
            If list : many input arrays.
            If dict : many input arrays, with keys specifying which parameter
                      of `Layer.evaluate` each array is associated with.
            (see examples below)
        output : str, list, or None; optional.
            Specifies name(s) for array output(s) of `Layer.evaluate`.
            If None : use default name specified by parent Model.
            If str  : same name for every output (incremented if multiple).
            If list : one name per output (length must match).
            (see examples below)
        parameters : nems.layers.base.Phi or None; optional.
            Specifies values for fittable parameters used by `Layer.evaluate`.
            If None : Phi returned by `Layer.initial_parameters`.
        priors : dict of Distributions or None; optional.
            Determines prior that each Layer parameter will sample values from.
            Keys must correspond to names of parameters, such that each
            Parameter utilizes `Parameter(name, prior=priors[name])`.
            If `None` : all parameters default to Normal(mean=zero, sd=one),
            where zero and one are appropriately shaped arrays of 0 and 1.
            Individual `None` entries in a `priors` dict result in the same
            behavior for those parameters.
        bounds : dict of 2-tuples or None; optional.
            Determines minimum and maximum values for fittable parameters. Keys
            must correspond to names of parameters, such that each Parameter
            utilizes `Parameter(name, bounds=bounds[name])`.
            If None : use defaults defined in `Layer.initial_parameters`.
        name : str or None; optional.
            A name for the Layer so that it can be referenced through the
            parent Model, in addition to integer indexing.

        Warnings
        --------

        See also
        --------
        nems.models.base.Model
        nems.layers.base.Phi
        nems.layers.base.Parameter

        Examples
        --------
        Subclasses that need to overwrite `__init__()` should specify new
        arguments (if any) in the method definition, followed by **kwargs, and
        invoke super().__init__(**kwargs) to ensure all required attributes are
        set correctly. While not strictly required, this is the easiest way to
        ensure Layers function properly within a Model.

        If `initial_parameters` needs access to new attributes, they should be
        set prior to invoking `super().__init__()`. New options that interact
        with base attributes (like `Layer.parameters` or `Layer.priors`) should
        be coded after invoking `super().__init__()`, to ensure the relevant
        attributes have been set.

        For example:
        >>> def __init__(self, new_arg1, new_kwarg2=None, **kwargs):
        ...     self.something_new = new_arg1
        ...     super().__init__(**kwargs)
        ...     self.do_something_to_priors(new_kwarg2)


        When specifying input for Layers, use:
        `None` to retrieve output of previous Model Layer (default).
        `'data_key'` to retrieve a single specific array.
        `['data_key1', ...]` to retrieve many arrays. This is preferred if
            order of arguments for `Layer.evaluate` is not important
        `{'arg1': 'data_key1', ...}` to map many arrays to specific arguments
            for `Layer.evaluate`.

        >>> data = {'stimulus': stim, 'pupil': pupil, 'state1': state1,
        ...         'state2': state2}
        >>> layers = [
        ...     WeightChannels(shape=(18,4), input='stimulus'),
        ...     FIR(shape=(4, 25)),
        ...     DoubleExponential(output='LN_output'),
        ...     Sum(input=['LN_output', 'state', 'pupil'],
        ...                output='summed_output'),
        ...     LinearWeighting(
        ...         input={
        ...             'pred': 'summed_output',
        ...             'pupil': 'pupil',
        ...             'other_states': ['state1', 'state2']
        ...             }
        ...         )
        ...     ]

        """

        # input/output should be string, list of strings,
        # dict of strings, or None
        self.input = input
        self.output = output

        self.initial_priors = priors
        self.initial_bounds = bounds
        if shape is not None: shape = tuple(shape)
        self.shape = shape
        # In the event of a name clash in a Model, an integer will be appended
        # to `Layer.name` to ensure that all Layer names are unique.
        # TODO: make name property instead and save _name, default_name here.
        #       (so that other code doesn't need to check for None name)
        self._name = name 
        self.default_name = f'{type(self).__name__}'
        self.model = None  # pointer to parent Model

        if parameters is None:
            parameters = self.initial_parameters()
        self.parameters = parameters 

        # Overwrite defaults set by `initial_parameters` if priors, bounds
        # kwargs were specified.
        if priors is not None:
            self.set_priors(**priors)
        if bounds is not None:
            self.set_bounds(**bounds)

        self.data_map = DataMap(self)

    @property
    def name(self):
        """Get Layer.name if specified, or `SubClass(shape=self.shape)."""
        name = self._name
        if name is None:
            name = self.default_name
        return name

    @property
    def parameter_count(self):
        """Total size of all Parameters.
        
        TODO: rename? (see note in Model.parameter_count)
        
        """

        return self.parameters.size

    @property
    def parameter_info(self):
        """Sizes of frozen, unfrozen and permanent parameters.
        
        Returns
        -------
        dict
            {'total': int, 'unfrozen': int, 'frozen': int, 'permanent': int}

        """

        total = self.parameters.size
        unfrozen = self.parameters.unfrozen_size
        frozen = total - unfrozen
        permanent = sum([p.is_permanent*p.size for p in self.parameters])

        # TODO: rename unfrozen -> trainable, frozen -> untrainable?
        return {'total': total, 'unfrozen': unfrozen, 'frozen': frozen,
                'permanent': permanent}

    def set_dtype(self, dtype):
        """Change dtype of all Parameter values in-place."""
        self.parameters.set_dtype(dtype)

    @layer('baseclass')
    def from_keyword(keyword):
        """Construct a Layer corresponding to a registered keyword.

        Each Layer subclass can (optionally) overwrite `from_keyword` to
        enable compatibility with the NEMS keyword system. This is a
        string-based shortcut system for quickly specifying a `Model` without
        needing to import individual Layer subclasses.

        To work correctly within this system, a `from_keyword` method must
        follow these requirements:
        1) The `@layer` decorator, imported from `nems.registry`, must be added
           above the method. The decorator must receive a single string as
           its argument, which serves as the keyword "head": the identifier for
           the relevant Layer subclass. This string must contain only letters
           (i.e. alphabetical characters - no numbers, punctuation, special
           characters, etc). Keyword heads are typically short and all
           lowercase, but this is not enforced.
        2) `from_keyword` must be a static method (i.e. it receives neither
           `cls` nor `self` as an implicit argument).
        3) `from_keyword` must accept a single argument. Within the keyword
           system, this argument will be a string of the form:
           `'{head}.{option1}.{option2}...'`
           where any number of options can be specified, separated by periods.
           Options can contain any* character other than hyphens or underscores,
           which are reserved for composing many keywords at once.
           Options for builtin NEMS layers mostly follow certain formatting
           norms for consistency, but these are not enforced:
           a) Use a lowercase 'x' between dimensions for shape:
                `shape=(3,4)`     -> '3x4'
           # TODO: replace examples for below with examples for layers
           #       (but they illustrate the format in the meantime)
           b) Scientific notation refers to negative exponents:
                `tolerance=0.001` -> 't1e3'
                `max_iter=1000`   -> 'i1000' or 'i1K'
           c) Boolean options use a single lowercase letter if possible:
                `normalize=True`  -> 'n'
           # TODO: any other formatting norms I missed?
           * Users should still be aware that escapes and other characters with
           python- or system-specific meanings should be avoided/used with care.
           Generally, sticking to alpha-numeric characters is safest.
        4) Return an instance of a Layer subclass.
        
        See also
        --------
        nems.registry
        nems.layers.weight_channels.WeightChannels.from_keyword

        Examples
        --------
        A minimal version of `from_keyword` for WeightChannels can be defined
        as follows:
        >>> class WeightChannels(Layer):
        >>>     def __init__(self, shape, **kwargs):
        >>>         self.shape = shape
        >>>         super().__init__(**kwargs)
        
        >>>     @layer('wtchans')
        >>>     @staticmethod
        >>>     def from_keyword(kw):
        ...         options = kw.split('.')
        ...         for op in options:
        ...             if ('x' in op) and (op[0].isdigit()):
        ...                 dims = op.split('x')
        ...                 shape = tuple([int(d) for d in dims])
        ...         # Raises UnboundLocalError if shape is never defined.
        ...         return WeightChannels(shape)

        >>> wtchans = WeightChannels.from_keyword('wtchans.18x2')
        >>> wtchans.shape
        (18, 2)
        
        Note: the actual definition referenced above uses `'wc'` as the keyword
        head, not `'wtchans'`. The key is changed for this example to avoid
        a name clash when testing.

        """
        return Layer()

    def initial_parameters(self):
        """Get initial values for `Layer.parameters`.
        
        Default usage is that `Layer.initial_parameters` will be invoked during
        construction to set `Layer.parameters` if `parameters is None`. Each
        Layer subclass should overwrite this method to initialize appropriate
        values for its parameters, and document the Layer's parameters in the
        overwritten docstring.

        Returns
        -------
        parameters : Phi

        See also
        --------
        nems.layers.weight_channels.WeightChannels.initial_parameters

        Examples
        --------
        A minimal version of `initial_parameters` for WeightChannels can be
        defined as follows:
        >>> class WeightChannels(Layer):
        >>>     def __init__(self, shape, **kwargs):
        >>>         self.shape = shape
        >>>         super().__init__(**kwargs)
        
        >>>     def initial_parameters(self):
        ...         coeffs = Parameter(name='coefficients', shape=self.shape)
        ...         return Phi(coeffs)

        >>> wc = WeightChannels(shape=(2, 1))
        >>> wc.parameters
        Parameter(name=coefficients, shape=(2, 1))
        .values:
        [[0.]
         [0.]]
        
        """
        return Phi()

    def reset_map(self):
        """Overwrite `Layer.data_map` with a fresh copy.
        
        This will remove information about outputs and any data-dependent
        information (like first input).
        
        """
        self.data_map = DataMap(self)

    def _evaluate(self, data, inplace_ok=False, dtype=np.float32,
                  debug_nans=False):
        """Get inputs from `data`, evaluate them, and update `Layer.data_map`.

        Parameters
        ----------
        data : dict
            See `Model.evaluate` for details on structure.
        inplace_ok : bool; default=False.
            If True, informs `Layer.evaluate` that it can safely overwrite
            `inputs` to reduce memory usage.
        dtype : type, default=np.float32.
            Output of `Layer.evaluate` will be cast to this type with
            `copy=False`.
        debug_nans : bool; default=False.
            If True, raise ValueError if output of `Layer.evaluate` contains
            NaN values.

        Returns
        -------
        args : list of ndarray
            Positional arguments for `Layer.evaluate`.
        kwargs : dict of ndarray
            Keyword arguments for `Layer.evaluate`.
        output : ndarray or list of ndarray
            Return value of `Layer.evaluate(*args, **kwargs)`.
        
        Notes
        -----
        This method is dependent on the state of `data`, so is best thought of
        as internal to `Model.evaluate` even though it technically does not
        rely on the state of a particular Model.

        See also
        --------
        nems.models.base.Model.evaluate
        nems.layers.base.map.DataMap

        """
        args, kwargs = self.data_map.get_inputs(data)
        self._inplace_ok = inplace_ok  # Set flag for `evaluate` behavior.
        output = self.evaluate(*args, **kwargs)
        self._inplace_ok = False       # Always reset to default of False.

        # Add singleton channel axis to each array if missing, and cast dtype
        # if necessary.
        if isinstance(output, (list, tuple)):
            output = [
                x[..., np.newaxis].astype(dtype, copy=False)
                if x.ndim == 1
                else x.astype(dtype, copy=False)
                for x in output
                ]
            if debug_nans and any([one_or_more_nan(o) for o in output]):
                raise ValueError(
                    f"Output of layer {self.name} contains NaNs."
                )
        else:
            output = output.astype(dtype, copy=False)
            if output.ndim == 1:
                output = output[..., np.newaxis]
            if debug_nans and one_or_more_nan(output):
                raise ValueError(
                    f"Output of layer {self.name} contains NaNs."
                )
        
        # Add output information to DataMap
        self.data_map.map_outputs(output)

        # args and kwargs are also returned for easier debugging
        return args, kwargs, output

    def evaluate(self, *args, **kwargs):  
        """Applies some mathematical operation to the argument(s).
        
        Each Layer subclass must overwrite this method. Any number of arguments
        is acceptable, but each should correspond to one name in `self.input`
        at runtime. An arbitrary number of return values is also allowed, and
        each should correspond to one name in `self.output`.
        
        Input and output names will be associated with arguments and return
        values, respectively, in list-order. If `self.input` is a dictionary,
        inputs will instead be mapped to specific arguments (where each
        key is the name of an argument, and each value is the name of an array
        in a `data` dictionary).

        Warnings
        --------
        Evaluate should never modify inputs in-place, as this could change the
        input to other Layers that expect the original data. Intermediate
        operations should always return copies. If a modified copy of the input
        is needed as standalone data, then it should be returned as a separate
        output (either by the same layer or a different one).
        
        Returns
        -------
        N-tuple of numpy.ndarray

        See also
        --------
        Layer.__init__

        Examples
        --------

        >>> class DummyLayer(Layer):
        >>>     def evaluate(self, x, y, z):
        ...         a = x + y
        ...         b = 2*z
        ...         return a, b
        
        >>> x = DummyLayer(input=['one', 'two', 'three'], output=['two', 'new'])
        >>> data = {'one': x, 'two': y, 'three': z}
        
        During fitting, `x.evaluate` would receive `(x, y, z)` as arguments
        (in that order) and return `(x+y, 2*z)`, resulting in:
        >>> data
        {'one': x, 'two': x+y, 'three': z, 'new': 2*z}

        """
        raise NotImplementedError(
            f'{type(self).__name__} has not defined `evaluate`.'
            )

    def description(self):
        """Short description of Layer's function.
        
        Defaults to `Layer.evaluate.__doc__` if not overwritten.

        Example
        -------
        def evaluate(self):
            '''A really long docstring with citations and notes other stuff.'''
            return a + np.exp(b-c)

        def description(self):
            return '''Implements a simple exponential: $a + e^{(b-c)}$'''

        """
        return self.evaluate.__doc__

    def as_tensorflow_layer(self, **kwargs):
        """Build an equivalent TensorFlow layer.

        Builds a subclass of NemsKerasLayer and returns an instance from:
        `SubClass(self, **kwargs)`.

        Parameters
        ----------
        kwargs : dict
            Keyword arguments for `NemsKerasLayer` or `tf.keras.layers.Layer`.

        Returns
        -------
        tf.keras.layers.Layer

        See also
        --------
        nems.tf.layer_tools.NemsKerasLayer

        Notes
        -----
        Subclasses should import TensorFlow modules inside this method, *not*
        at the top of the module. This is done to ensure that TensorFlow does
        not need to be installed unless it is being used.

        """
        raise NotImplementedError(
            f'{self.__class__} has not defined a Tensorflow implementation.'
            )

    def as_tf(self):
        """Alias for `as_tensorflow_layer`. This should not be overwritten."""
        return self.as_tensorflow_layer()

    def get_parameter_values(self, *parameter_keys, as_dict=False):
        """Get all parameter values, formatted as a list or dict of arrays.

        Parameters
        ----------
        parameter_keys : N-tuple of strings
            Keys indicating which parameters to get values for. If no keys are
            specified, all parameter values will be fetched.
        as_dict : bool
            If True, return a dict instead of a list with format
            `{parameter_key : Layer.parameters[parameter_key].values}`.

        Returns
        -------
        values : list or dict
        
        See also
        --------
        Phi.get_values
        
        """
        if parameter_keys == ():
            parameter_keys = self.parameters.keys()
        values = self.parameters.get_values(*parameter_keys)
        if as_dict:
            values = {k: v for k, v in zip(parameter_keys, values)}
        return values

    def set_parameter_values(self, *parameter_dict, ignore_bounds=False,
                             **parameter_kwargs):
        """Set new parameter values from key, value pairs.

        Parameters
        ----------
        parameter_dict : dict; optional
            Dictionary containing key-value pairs, where each key is the name
            of a parameter and each value is the new value for that parameter.
        ignore_bounds : bool
            If True, ignore `Parameter.bounds` when setting new values
            (not usually recommended, but useful for testing).
        parameter_kwargs : key-value pairs.
            As `parameter_dict`.

        """
        self.parameters.update(*parameter_dict, ignore_bounds=ignore_bounds,
                               **parameter_kwargs)

    def get_parameter_vector(self, as_list=True):
        """Get all parameter values, formatted as a single vector.
        
        Parameters
        ----------
        as_list : bool
            If True, returns a list instead of ndarray.

        Returns
        -------
        list or ndarray

        See also
        --------
        Phi.get_vector

        """
        return self.parameters.get_vector(as_list=as_list)

    def get_parameter_from_index(self, index):
        """Get reference to Parameter corresponding to vector index."""
        return self.parameters.get_parameter_from_index(index)

    def set_parameter_vector(self, vector, ignore_checks=False):
        """Set parameter values with a single vector.

        Parameters
        ----------
        vector : ndarray or list
            New parameter vector. Size must match the total number of flattened
            parameter values.
        ignore_checks : bool
            If True, set new values without checking size or bounds.
            (intended as a minor optimization for the scipy fitter).

        See also
        --------
        Phi.set_vector
        
        """
        self.parameters.set_vector(vector, ignore_checks=ignore_checks)

    def get_bounds_vector(self, none_for_inf=True):
        """Get all parameter bounds, formatted as a list of 2-tuples.

        Parameters
        ----------
        none_for_inf : bool
            If True, +/- np.inf is replaced with None
            (for scipy-compatible bounds).

        Returns
        -------
        list of 2-tuple

        See also
        --------
        Phi.get_bounds

        """
        return self.parameters.get_bounds_vector(none_for_inf=none_for_inf)

    @property
    def bounds(self):
        """Get all parameter bounds, as a dict with one key per Parameter."""
        return self.parameters.bounds
    
    def set_bounds(self, *parameter_dict, **parameter_kwargs):
        """Set all parameter bounds fromm key-value pairs.
        
        Warnings
        --------
        This can force the current `Parameter.values` out of bounds. To ensure
        valid values, a subsequent `Parameter.sample` or `Parameter.mean` is
        recommended.

        """
        self.parameters.set_bounds(*parameter_dict, **parameter_kwargs)

    @property
    def priors(self):
        """Get all parameter priors, as a dict with one key per Parameter."""
        return self.parameters.priors

    def set_priors(self, *parameter_dict, **parameter_kwargs):
        """Set all parameter bounds fromm key-value pairs.
        
        Warnings
        --------
        This can force the current `Parameter.values` to be outside the range
        of `Parameter.prior`. To ensure valid values, a subsequent
        `Parameter.sample` or `Parameter.mean` is recommended.

        """
        self.parameters.set_priors(*parameter_dict, **parameter_kwargs)

    def sample_from_priors(self, inplace=True, as_vector=False):
        """Get or set new parameter values by sampling from priors.
        
        Parameters
        ----------
        inplace : bool, default=True
            If True, sampled values will be used to update each Parameter
            (and, in turn, Phi._array) inplace. Otherwise, the
            sampled values will be returned without changing current values.
        as_vector : bool, default=False
            If True, return sampled values as a flattened vector instead of a
            list of arrays.

        Returns
        -------
        samples : ndarray or list of ndarray

        See also
        --------
        nems.layers.base.Phi.sample

        """
        return self.parameters.sample(inplace=inplace, as_vector=as_vector)

    def mean_of_priors(self, inplace=True, as_vector=False):
        """Get, or set parameter values to, mean of priors.
        
        Parameters
        ----------
        inplace : bool, default=True
            If True, mean values will be used to update each Parameter
            (and, in turn, `Phi._array`) inplace. Otherwise, means
            will be returned without changing current values.
        as_vector : bool, default=False
            If True, return means as a flattened vector instead of a
            list of arrays.

        Returns
        -------
        means : ndarray or list of ndarray

        See also
        --------
        nems.layers.base.Phi.mean

        """
        return self.parameters.mean(inplace=inplace, as_vector=as_vector)

    def freeze_parameters(self, *parameter_keys):
        """Use parameter values for evaluation only, do not optimize.

        Parameters
        ----------
        parameter_keys : N-tuple of str
            Each key must match the name of a Parameter in `Layer.parameters`.
            If no keys are specified, all parameters will be frozen.
        
        See also
        --------
        Phi.freeze_parameters
        Parameter.freeze
        
        """
        self.parameters.freeze_parameters(*parameter_keys)

    def unfreeze_parameters(self, *parameter_keys):
        """Make parameter values optimizable.

        Parameters
        ----------
        parameter_keys : N-tuple of str
            Each key must match the name of a Parameter in `Layer.parameters`.
            If no keys are specified, all parameters will be unfrozen.
        
        See also
        --------
        Phi.unfreeze_parameters
        Parameter.unfreeze

        """
        self.parameters.unfreeze_parameters(*parameter_keys)

    def set_permanent_values(self, *parameter_dict, **parameter_kwargs):
        """Set parameters to fixed values. The parameters will not unfreeze."""
        self.parameters.set_permanent_values(*parameter_dict, **parameter_kwargs)

    def set_index(self, i, new_index='initial'):
        """Change which set of parameter values is referenced.

        Intended for use with jackknifing or other procedures that fit multiple
        iterations of a single model. Rather than using many copies of a full
        Model object, each Phi object tracks copies of the underlying vector of
        parameter values.

        Parameters
        ----------
        i : int
            New index for `Phi._array`. If `i >= len(Phi._array)`, then new
            vectors will be appended until `Phi._array` is sufficiently large.
        new_index : str or None, default='initial'
            Determines how new vectors are generated if `i` is out of range.
            If `'sample'`   : invoke `Phi.sample(inplace=False)`.
            Elif `'mean'`   : invoke `Phi.mean(inplace=False)`.
            Elif `'initial'`: set to `[p.initial_value for p in <Parameters>]`.
            Elif `'copy'`   : copy current `Phi.get_vector()`.
            Elif `None`     : raise IndexError instead of adding new vectors.

        See also
        --------
        nems.models.base.Model.set_index

        """
        self.parameters.set_index(i, new_index=new_index)

    def plot(self, output, fig=None, ax=None, **plot_kwargs):
        """Alias for `nems.visualization.model.plot_layer`.

        By default, layer output will be represented by a single 2D line plot
        with time on the x-axis.

        Parameters
        ----------
        output : list of ndarray
            Data returned by `Layer.evaluate`.
        fig : matplotlib.figure.Figure; optional.
            The figure on which the plot will be rendered. If not provided, a
            new figure will be generated.
        plot_kwargs : dict
            Additional keyword arguments to be supplied to
            `matplotlib.axes.Axes.plot`.

        Returns
        -------
        fig : matplotlib.figure.Figure

        See also
        --------
        nems.visualization.model.plot_model
        nems.visualization.model.plot_layer

        Notes
        -----
        Subclasses can overwrite this method to specify a custom plot.
        To be compatible with `nems.tools.plotting.plot_model`, the overwritten
        method must accept `output` as the first argument and `fig` as a keyword
        argument. The plot must be rendered on the provided Figure (unless
        `fig is None`), but `output` does not necessarily need to be used (for
        example, when visualizing model parameters).

        Examples
        --------
        >>> class MyNewLayer(Layer):
        >>>     def evaluate(self, *inputs):
        >>>         return [2*x for x in inputs]
        >>>     def plot(self, output, fig=None):
        >>>         plot_data = some_special_routine(output)
        >>>         if fig is None:
        >>>             fig = plt.figure()
        >>>         ax = fig.subplots(1,1)
        >>>         ax.plot(plot_data)
        >>>         return fig
        >>> # Use evaluate to generate output.
        >>> layer = MyNewLayer()
        >>> output = layer.evaluate(np.random.rand(100, 1))
        >>> layer.plot(output)
        
        """
        return plot_layer(output, fig=fig, ax=ax, **plot_kwargs)

    @property
    def plot_kwargs(self):
        """Get default keyword arguments for `Layer.plot`.
        
        These kwargs will be used by `nems.visualization.model.plot_model`.

        Returns
        -------
        dict
            If `Layer.plot` has *not* been overwritten, each key must correspond
            to one keyword argument for `matplotlib.axes.Axes.plot`, such that
            `ax.plot(..., **Layer.plot_kwargs)` is valid.

            If `Layer.plot` *has* been overwritten, each key must correspond to
            one keyword argument for `Layer.plot`, such that
            `Layer.plot(..., **Layer.plot_kwargs)` is valid.

        See also
        --------
        nems.visualization.model.plot_model
        
        """
        return {}
    
    @property
    def plot_options(self):
        """Get default plot options for `Layer.plot`.

        These options will be used by `nems.visualization.model.plot_model`.

        Returns
        -------
        dict
            Dictionary keys should correspond to a subset of the keys in
            `nems.visualization.model._DEFAULT_PLOT_OPTIONS`. See function
            `set_plot_options` in the same module for behavior.
        
        See also
        --------
        nems.visualization.model.plot_model
        nems.visualization.model.set_plot_options

        Notes
        -----
        To prevent all default plot options from being used (for example, to
        hard-code specific options in an overwritten `Layer.plot` method), set
        `{'skip_plot_options': True}` in the returned dictionary.
        
        """
        return {}

    # Passthrough *some* of the dict-like interface for Layer.parameters
    # NOTE: 'get' operations through this interface return Parameter instances,
    #       not arrays. For array format, reference Parameter.values or use
    #       `Layer.get_parameter_values`.
    def __getitem__(self, key):
        """Get Parameter (not Parameter.values)."""
        return self.parameters[key]
    
    def get(self, key, default=None):
        """Get Parameter (not Parameter.values)."""
        return self.parameters.get(key, default=default)

    def __setitem__(self, key, val):
        """Set Parameter.values (not Parameter itself)."""
        self.parameters[key] = val

    def __iter__(self):
        """Return an iter-wrapped singleton list containing this Layer.

        Supports compatibility for iterating over Model layers when single
        Layers are returned, without the need to repeatedly add `[0]` to the
        end of indexing statements.
        
        NOTE: This does *not* iterate over the Layer's parameters, and is not
              part of the passed-through dict interface for `Layer.Phi`. To
              iterate over parameters, refer iterators directly to
              `Layer.parameters`.
              E.x: `for p in Layer.parameters` instead of `for p in Layer`.

        TODO: This feels kludgy and I don't like it, but it allows for some
              nice syntax for navigating Models.
        
        """
        return iter([self])

    def __str__(self):
        header, equal_break = self._repr_helper()
        string = header + equal_break + "\n"
        if self.state_arg is not None:
            string += f".state_arg:  {self.state_arg}\n"
        string += ".parameters:\n\n"
        p_string = str(self.parameters)
        string += p_string
        if p_string != '':
            # Only add an end-cap equal_break if there was at least one
            # parameter to show
            string += equal_break
        return string

    def __repr__(self):
        header, equal_break = self._repr_helper()
        string = header + equal_break + "\n"
        p_string = self.parameters.__repr__()
        string += p_string
        if p_string != '':
            # Only add an end-cap equal_break if there was at least one
            # parameter to show
            string += equal_break
        return string

    def _repr_helper(self):
        layer_dict = Layer().__dict__
        self_dict = self.__dict__
        # Get attributes that are not part of base class, unless they start
        # with an underscore, then convert to string with "k=v" format
        self_only = ", ".join([f"{k}={v}" for k, v in self_dict.items()
                               if (k not in layer_dict) and not
                               (k.startswith('_'))])
        if self_only != "":
            # Prepend comma to separate from shape
            self_only = ", " + self_only
        header = f"{type(self).__name__}(shape={self.shape}{self_only})\n"
        equal_break = "="*32

        return header, equal_break


    # Add compatibility for saving to .json
    def to_json(self):
        """Encode a Layer as a dictionary.

        This (base class) method encodes all attributes that are common to
        all Layers. Subclasses that need to save additional kwargs or attributes
        should overwrite `to_json`, but invoke `Layer.to_json` within the new
        method as a starting point (see example below).
        
        See also
        --------
        'Layer.from_json`
        `nems.tools.json`
        `nems.layers.weight_channels.WeightChannels.to_json`

        Examples
        --------
        >>> class DummyLayer(Layer):
        >>>     def __init__(self, **kwargs):
        ...         super().__init__(**kwargs)
        ...         important_attr = None
        >>>     def update_important_attr(self, *args):
        ...         important_attr = do_stuff(args)
        ...     # ...
        ...     # Want to preserve the state of `important_attr` in encoding.
        >>>     def to_json(self):
        ...         data = Layer.to_json(self)
        ...         data['attributes'].update(
        ...             important_attr=self.important_attr
        ...             )
        ...         return data

        """
        data = {
            'kwargs': {
                'input': self.input, 'output': self.output,
                'parameters': self.parameters, 'priors': self.priors,
                'bounds': self.bounds, 'name': self.name, 'shape': self.shape,
            },
            'attributes': {},
            'class_name': type(self).__name__
            }

        return data

    @classmethod
    def from_json(cls, json):
        """Decode a Layer from a dictionary.

        Parameters
        ----------
        json : dict
            json data encoded by `Layer.to_json`.

        Returns
        -------
        layer : Layer or subclass
        
        See also
        --------
        `Layer.to_json`
        `nems.tools.json`.

        """
        if json['class_name'] == 'Layer':
            subclass = Layer
        else:
            subclass = cls.subclasses[json['class_name']]
        if subclass.from_json.__qualname__ != Layer.from_json.__qualname__:
            # Subclass has overwritten `from_json`, use that method instead.
            layer = subclass.from_json()
        else:
            layer = subclass(**json['kwargs'])
            for k, v in json['attributes'].items():
                setattr(layer, k, v)

        return layer

    def __eq__(self, other):
        # Must be the same Layer subclass and have the same parameters.
        if type(self) is type(other):
            return self.parameters == other.parameters
        else:
            return NotImplemented



import scipy.optimize

from .cost import get_cost
from ..base import Backend, FitResults


class SciPyBackend(Backend):
    """Default backend: just passthrough NEMS model methods except fit."""
    def _build(self, data=None, eval_kwargs=None, **backend_options):
        """Duplicate reference to `SciPyBackend.nems_model`."""
        # This is the default backend, so there's nothing to build.
        # SciPyBackend.model and .nems_model are the same.
        return self.nems_model

    def _fit(self, data, eval_kwargs=None, cost_function='nmse',
             epochs=1, log_spacing=5, **fitter_options):
        """Fit a Model using `scipy.optimize.minimize`.

        Parameters
        ----------
        data : DataSet
            Model inputs, outputs and targets.
        eval_kwargs : dict
            Keyword arguments for `nems_model.evaluate`.
        cost_function : str or function; default='nmse'
            Specifies which metric to use for computing error while fitting.
            Default is mean squared error normalized by the standard deviation
            of the target.
            If str      : Replace with `get_cost(str)`.
            If function : Use this function to compute errors. Should accept
                          two array arguments and return float.
        epochs : int; default=1.
            Number of "outer loops" to repeat optimization over. Note that this
            is redundant with `fitter_options={'options': {'maxiter': N}}` if
            data contains a single batch (or no batches).
        log_spacing : int; default=5.
            Log progress of fitter every `log_spacing` iterations.
            NOTE: this is the number of iterations, not the number of cost
            function evaluations (of which there may be many per iteration).

        Returns
        -------
        FitResults

        Notes
        -----
        For this Backend, additional **fitter_options are captured because they
        need to be passed directly to `scipy.optimize.minimize`.

        Currently this fitter is designed to work with `method='L-BFGS-B'`.
        Other methods can be used by specifying a 'method' key in
        `fitter_options`, but syntax compatibility with other minimize methods
        has not been tested.

        See also
        --------
        nems.models.base.Model.fit

        """

        if eval_kwargs is None: eval_kwargs = {}
        eval_kwargs['use_existing_maps'] = True
        eval_kwargs['as_dataset'] = True
        batch_size = eval_kwargs.get('batch_size', 0)

        if isinstance(cost_function, str):
            # Convert string reference to loss function
            cost_function = get_cost(cost_function)
        wrapper = _FitWrapper(
            cost_function, self.nems_model, eval_kwargs, log_spacing
            )
        
        # TODO: check tolerance change between epochs.
        for ep in range(epochs):
            print(f"Epoch {ep}")
            print("="*30)
            if batch_size == 0:
                # Don't need to mess with batching.
                _data = data.copy()
                fit_result = wrapper.get_fit_result(_data, **fitter_options)
            else:
                # TODO: maybe better to only generate batches once? Would have
                #       to load them all in a list, but they should be views
                #       anyway.
                batch_generator = self.nems_model.generate_batch_data(
                    data, **eval_kwargs
                    )
                for i, _data in enumerate(batch_generator):
                    print(" "*4 + f"Batch {i}")
                    print(" "*4 + "-"*26)
                    fit_result = wrapper.get_fit_result(_data, **fitter_options)

        print(
            f"Fit successful: {fit_result.success}\n"
            f"Status: {fit_result.status}\n"
            f"Message: {fit_result.message}"
        )
        self.model.set_parameter_vector(fit_result.x)
        
        initial_parameters = wrapper.initial_parameters
        final_parameters = wrapper.model.get_parameter_vector()
        initial_error = wrapper(initial_parameters)
        final_error = wrapper(final_parameters)
        nems_fit_results = FitResults(
            initial_parameters, final_parameters, initial_error, final_error,
            backend_name='SciPy', scipy_fit_result=fit_result
        )

        return nems_fit_results

    def predict(self, input, **eval_kwargs):
        """Alias for `SciPyBackend.nems_model.predict`."""
        self.nems_model.predict(input, **eval_kwargs)


class _FitWrapper:

    def __init__(self, fn, model, eval_kwargs, log_spacing):
        """Internal for `SciPyBackend._fit`."""
        self.fn = fn
        try:
            self.name = fn.__name__
        except AttributeError:
            self.name = 'none'
        self.model = model
        self.initial_parameters = model.get_parameter_vector(as_list=True)
        self.bounds = model.get_bounds_vector(none_for_inf=True)
        self.eval_kwargs = eval_kwargs
        self.log_spacing = log_spacing
        self.data = None
        self.iteration = 0

    def __call__(self, vector):
        """Update model parameters with new vector, compute cost."""
        self.model.set_parameter_vector(vector, ignore_checks=True)
        evaluated_data = self.model.evaluate(self.data, **self.eval_kwargs)
        self.data = evaluated_data
        cost = self.compute_cost()
        return cost

    # TODO: better callback scheme, use logging
    def callback(self, vector):
        """Print error after most recent iteration based on `log_spacing`."""
        if self.iteration % self.log_spacing == 0:
            # Shouldn't need to set parameter vector, that should have
            # been done by the optimization iteration.
            cost = self.__call__(vector)
            print(" "*8 + f"Iteration {self.iteration},"
                    f" error is: {cost:.8f}...")
        self.iteration += 1

    def compute_cost(self):
        """Compute cost given current Model parameters."""
        prediction_list, target_list = self._get_arrays()
        if (len(prediction_list) == 1) and (len(target_list) == 1):
            cost = self.fn(prediction_list[0], target_list[0])
        else:
            # Dict keys are not stored in a guaranteed order, so can't expect
            # .values() to match up even if the lengths are the same. Need to
            # provide a separate mapping of {'pred_key' -> 'target_key'}
            # (and do something different instead of just getting lists).
            raise NotImplementedError(
                "TODO: SciPy cost function for multiple predictions/targets."
                )

        return cost

    def _get_arrays(self):
        """Retrieve outputs and targets from `data` as lists of arrays."""
        prediction = self.data.outputs
        target = self.data.targets

        if len(prediction) == 0:
            # No predictions, error
            raise ValueError(f"{self.name}: No predictions found in data.")
        else:
            predictions = list(prediction.values())
        
        if len(target) == 0:
            # No target, error
            raise ValueError(f"{self.name}: No targets found in data.")
        else:
            targets = list(target.values())

        return predictions, targets

    def get_fit_result(self, data, **fitter_options):
        """Process one optimization epoch."""
        self.data = data
        self.iteration = 0

        if 'method' not in fitter_options:
            fitter_options['method'] = 'L-BFGS-B'
        fit_result = scipy.optimize.minimize(
            self, self.initial_parameters, bounds=self.bounds,
            callback=self.callback, **fitter_options
            )

        return fit_result



"""Cost functions for SciPyBackend."""

import numpy as np

from nems.tools.lookup import FindCallable


# TODO: Need to have normalization in-place for multi-neuron data
#       (or multi-batch?), otherwise some components will get weighted
#       more heavily. Should be a pre-processing step.

# TODO: I don't love having different argument names for mse vs nmse, but I
#       wanted to make it clear that the ordering doesn't matter for the former
#       (and does for the latter). Similar for other modules in this directory.
def mse(x, y):
    """Compute the mean squared error (MSE) between arrays x and y.

    Parameters
    ----------
    x, y : np.ndarray
        Arrays must have the same shape. Most commonly, these will be a
        model output (prediction) and a recorded response (target).

    Returns
    -------
    mse : float

    Examples
    --------
    >>> prediction = model.predict(data)
    >>> target = data['response']
    >>> error = mse(output, target)

    >>> model.score()

    Notes
    -----
    This implementation is compatible with both numeric (i.e., Numpy)
    and symbolic (i.e., Theano, TensorFlow) computation.

    """
    squared_errors = (x - y)**2
    return np.mean(squared_errors)


def nmse(prediction, target):
    """Compute MSE normalized by the standard deviation of target.

    Because this metric is more computationally expensive than MSE, but is
    otherwise equivalent, we suggest using `mse` for fitting and `nmse` as a
    post-fit performance metric only.

    Parameters
    ----------
    prediction : ndarray
    target : ndarray
        Shape must match prediction. MSE will be divided by the
        standard deviation of this array.

    Returns
    -------
    normalized_error : float

    See also
    --------
    .mse
    
    """
    std_of_target = np.std(target)
    error = np.sqrt(mse(prediction, target))

    if std_of_target == 0:
        # This means the target is a constant. Return raw mse, so that fits will converge
        normalized_error = error
    else:
        normalized_error = error / std_of_target
    return normalized_error


# Add nickname here if desired string name doesn't match
# the name of the function.
cost_nicknames = {}
get_cost = FindCallable({**globals(), **cost_nicknames}, header='Cost')
    


import requests
import pathlib

import numpy as np

from nems.tools.arrays import apply_to_dict


# E.g. "/path/to/NEMS/nems/tools/demo_data/saved_data/"
basepath = pathlib.Path(__file__).parent / 'saved_data'
demo_cell = 'TAR010c-18-2.npz'


def download_data(url, filepath):
    r = requests.get(url, stream=True)
    with open(filepath, 'wb') as f:
        for chunk in r.iter_content(chunk_size=1024):
            if chunk:
                f.write(chunk)


# TODO: Other download options, or just the one file?
def download_demo(overwrite=False):
    prefix = 'https://s3-us-west-2.amazonaws.com/nemspublic/sample_data/'
    url = prefix + demo_cell
    filepath = basepath / demo_cell

    if filepath.is_file() and not overwrite:
        raise ValueError(
            "Demo data already exists at filepath:\n"
            f"{filepath}\n"
            "To overwrite data, use `download_demo(overwrite=True)`."
        )

    print(f"Downloading NEMS demo data from:\n{url}\n...\n")
    download_data(url, filepath)
    print(
        "NEMS demo data has been successfully downloaded to:\n"
        f"{filepath}\n\n"
        "Data can be loaded using:\n"
        "`training_dict, test_dict = nems.load_demo()`\n"
        "and indexed with:\n"
        "`training_dict['spectrogram']  # 100hz spectrogram\n"
        "`training_dict['response']     # 100hz PSTH / firing rate"
    )


# TODO: other options?
# TODO: batched version?
def load_demo(tutorial_subset=False):
    filepath = basepath / demo_cell
    if not filepath.is_file():
        raise ValueError(
            "NEMS demo data does not exist at path:\n"
            f"{filepath}\n"
            "First use `nems.download_demo()`, then try `load_demo()` again."
        )
    else:
        # TODO: just save the file this way instead?
        #       possibly rename it as well, to something like:
        #       "a1_100hz_linear.npz"
        with np.load(filepath) as data_dict:        
            training_dict = {
                'spectrogram': data_dict['X_est'],
                'response': data_dict['Y_est']
            }
            test_dict = {
                'spectrogram': data_dict['X_val'],
                'response': data_dict['Y_val']
            }

    if tutorial_subset:
        # TODO: 2900 is hard-coded for TAR010c-18-2, maybe make this its own
        #       file instead?
        select_subset = lambda a: a[:2900]  # first 5 stimuli
        training_dict = apply_to_dict(select_subset, training_dict)
        test_dict = apply_to_dict(select_subset, training_dict)

    # TODO: how to report meta information? maybe just print here?

    return training_dict, test_dict
